# @author: Matthew Owen
# @date:   5/24/2010
# @file:   hooks.zelig.R
#
# Rudimentary Hook System for Zelig
# executes code based on context


# @name:  name of the function being hooked
# @hook:  the actual hook function
# @when:  the keyword that triggers the hook
# return: None.
add.hook <- function (name, hook, when="") {
  # Error-checking
  if (!is.character(name))
    stop("Error: names must be a character data-type.")

  if (!(is.character(hook) || is.function(hook)))
    stop("Error: hook must be a character or a function.")

  if (!is.character(when))
    stop("Error: when must be a character data-type.")

  if (!(exists("gl.hooks", mode="list") &&
        exists("gl.hooks", envir=.GlobalEnv, mode="list")))
    assign("gl.hooks", list(), envir=.GlobalEnv)

  # Assign entry in the next largest environment (or .GlobalEnvir)
  gl.hooks[[when]][[name]] <<- c(gl.hooks[[when]][[name]], hook)
}


# @fcall: the function call that will hold the arguments to be passed
#         into the hook
# @when:  the entry to look up in the hook list
# @env:   optional argument to specify where the hooks are kept
# return: None.
# notes:  This function should only be run by the Zelig software
run.hook <- function (fcall=sys.call(which=-1), when="", env=NULL) {
  # Error-checking
  if (!(is.call(fcall) && is.character(when))) {
    warning("Warning: Invalid parameters passed into hook function")
    return()
  }

  # the list of hooks
  if (!exists("gl.hooks", mode="list"))
    return()

  # get the list of hooks
  hook.list <- gl.hooks[[when]]

  # function name
  fname <- as.character(fcall[[1]])

  # Execute the hooks
  for (li in hook.list[[fname]])
    do.call(li, as.list(fcall[-1]))
}


# @f:      a character string specifying the function to be extended
# @hook:   a pointer to a function (or string).  The hook to be added
# @params: <disabled> a set of parameters to be passed to the hook in
#          addition to the regular function call
# @envir:  the environment where the function will be found
# return:  a pointer to the new, extended function
# usage:
# sin <- add.hook.begin.priv(sin, function (x) cat("res ="))
add.hook.begin <- function(f,
                           hook,
                           params=list(),
                           envir=parent.frame()
                           ) {
  # catch bad input
  if (!(is.character(f) || is.function(f)))
    stop("Error: f must be a string.")

  if (length(f) != 1)
    stop("Error: f must have a length of 1.")

  
  # store memory address of old function
  old <- get(f, envir=envir)

  # return new anonymous function
  function (...) {
    # get parameters (in local memory)
    mf <- match.call(expand.dots=T)
    args <- as.list(mf[-1])

    # execute hook
    do.call(hook, args)

    # execute original function
    do.call(old, args)
  }
}

# @f:      a list of character string functions
# @hook:   a pointer to a function
# @params: a list of parameters that are passed
#          into the new function (currently disabled)
# @envir:  an environment object
# return:  None.
# notes:   this function is more clever than the other hook methods,
#          however, it probably runs a bit slower and 
apply.hooks <- function(f,
                        hook,
                        params=list(),
                        envir=parent.frame()
                        ) {
  # catch invalid input
  if (!all(is.character(f)))
    stop("Error: all elements of f must be character-types.")


  # assign hook to the function list
  for (func in f)
    assign(func,
           add.hook.begin(func, hook,
                          params, envir=envir),
           envir=envir
           )
}
